import unittest as u
import os
from lose import LOSE
import lose
import numpy as np
import tables as t

v = [int(i) for i in lose.__version__.split('.')]

class Tests(u.TestCase):
	def setUp(self):
		if os.path.isfile('./temp.h5'):
			os.unlink('./temp.h5')

		self.l = LOSE('./temp.h5')

	def test_mk_group_valid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		self.l.newGroup(fmode='w', x=(15, 5), y=(2,))

	def test_mk_group_valid2(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		self.l.newGroup(fmode='a', y=(2,))
		self.l.newGroup(fmode='a', x=(15, 5))

	def test_mk_group_invalid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		with self.assertRaises(ValueError):
			self.l.newGroup(fmode='t', y=(2,))

	def test_save_valid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		check = True
		E = None

		self.l.newGroup(fmode='w', x=(25, 4), y=(2,))

		self.l.save(x=np.zeros((10, 25, 4)), y=np.zeros((2, 2)))
		self.l.save(x=np.zeros((15, 25, 4)), y=np.zeros((5, 2)))
		self.l.save(x=np.zeros((50, 25, 4)), y=np.zeros((8, 2)))

	def test_save_invalid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		self.l.newGroup(fmode='w', x=(25, 4), y=(2,))

		with self.assertRaises(ValueError):
			self.l.save(x=np.zeros((25, 4)), y=np.zeros((2, 2)))

	def test_save_invalid2(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		self.l.newGroup(fmode='w', x=(25, 4), y=(2,))

		with self.assertRaises(ValueError):
			self.l.save(x=np.zeros((10, 25, 4)), y=np.zeros((2, 5)))

	def test_save_invalid3(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		self.l.newGroup(fmode='w', x=(25, 4), y=(2,))

		with self.assertRaises(TypeError):
			self.l.save(x='lul')

	def test_load_valid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		a, b = self.l.load('x', 'y')

		self.assertEqual(np.all(a==X), np.all(b==Y), 'should be equal')

	@u.skipIf(v < [0, 6, 0], 'unsupported version')
	def test_load_valid2(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		a, b = self.l.load('x', 'y', batch_obj=':5')

		self.assertEqual(np.all(a==X[:5]), np.all(b==Y[:5]), 'should be equal')

	@u.skipIf(v > [0, 5, 0], 'version 0.5.0 and below only')
	def test_load_valid2_old(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		a, b = self.l.load('x', 'y', batch_obj='[:5]')

		self.assertEqual(np.all(a==X[:5]), np.all(b==Y[:5]), 'should be equal')

	def test_load_invalid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		with self.assertRaises(TypeError):
			a, b = self.l.load('x', 'y', batch_obj=None)

	def test_load_invalid2(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		with self.assertRaises(t.exceptions.NoSuchNodeError):
			a, b = self.l.load('z', 'g', batch_obj=None)

	@u.skipIf(v < [0, 5, 0], 'version 0.5 and up only')
	def test_rename_group_valid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		self.l.renameGroup(x='z', y='g')

		a, b = self.l.load('z', 'g')

		self.assertEqual(np.all(X == a), np.all(Y == b), 'should be equal')

	@u.skipIf(v < [0, 5, 0], 'version 0.5 and up only')
	def test_rename_group_invalid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		with self.assertRaises(t.exceptions.NoSuchNodeError):
			self.l.renameGroup(g='x')

	@u.skipIf(v < [0, 4, 5], 'version 0.4.5 and up only')
	def test_rm_group_valid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		self.l.removeGroup('x', 'y')

		with self.assertRaises(t.exceptions.NoSuchNodeError):
			a, b = self.l.load('x', 'y')

	@u.skipIf(v < [0, 4, 5], 'version 0.4.5 and up only')
	def test_rm_group_invalid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		with self.assertRaises(t.exceptions.NoSuchNodeError):
			self.l.removeGroup('x', 'y')

	@u.skipIf(v < [0, 6, 0], 'version 0.6 and up only')
	def test_getShapes_valid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		a, b = self.l.getShapes('x', 'y')

		self.assertEqual(a, X.shape, 'should be equal')
		self.assertEqual(b, Y.shape, 'should be equal')

	@u.skipIf(v > [0, 4, 5], 'version 0.4.5 and below only')
	def test_getShapes_valid_old(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=(0, X.shape[1:]), y=(0, Y.shape[1:]))
		self.l.save(x=X, y=Y)

		a = self.l.get_hape('x')
		b = self.l.get_hape('y')

		self.assertEqual(a, X.shape, 'should be equal')
		self.assertEqual(b, Y.shape, 'should be equal')

	@u.skipIf(v < [0, 4, 5], 'version 0.4.5 and up only')
	def test_getShapes_invalid(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)

		X = np.zeros((10, 5, 10))
		Y = np.zeros((10, 5))

		self.l.newGroup(fmode='w', x=X.shape[1:], y=Y.shape[1:])
		self.l.save(x=X, y=Y)

		with self.assertRaises(t.exceptions.NoSuchNodeError):
			a = self.l.getShape('g')

	# def test_generator_valid(self):
	# 	if os.path.isfile(self.l.fname):
	# 		os.unlink(self.l.fname)

		

	def tearDown(self):
		if os.path.isfile(self.l.fname):
			os.unlink(self.l.fname)


import numpy as np
import tables as t
import os
from contextlib import contextmanager


class LOSE:
	def __init__(self, fname=None):
		self.fname = fname
		self.atom = t.Float32Atom()

		self.batch_size = 1

		self.iterItems = None
		self.iterOutput = None
		self.loopforever = False
		self.limit = None
		self.shuffle = False
		self.mask_callback = None

		self._a = []
		self._b = []
		self._useCallback = False

	def __repr__(self):
		messsage = '<lose hdf5 data handler, fname={}, atom={}>'.format(self.fname, self.atom)
		messsage += '\ngenerator parameters: iterItems={}, iterOutput={}, batch_size={}, limit={}, loopforever={}, shuffle={}'.format(self.iterItems, self.iterOutput, self.batch_size, self.limit, self.loopforever, self.shuffle)
		if self.fname is not None:
			try:
				with t.open_file(self.fname, mode='r') as f:
					messsage += '\nhdf5 file structure:\n{}'.format(f.__repr__())

			except Exception as e:
				messsage += '\nfailed to open file at \'{}\':{}, make sure it\'s a not corrupted hdf5 file and is a real file'.format(self.fname, e)

		return messsage

	def __str__(self):
		messsage = '<lose hdf5 data handler, fname={}, atom={}>'.format(self.fname, self.atom)
		if self.fname is not None:
			try:
				with t.open_file(self.fname, mode='r') as f:
					messsage += '\nhdf5 file structure:\n{}'.format(f)

			except Exception as e:
				messsage += '\nfailed to open file at \'{}\':{}, make sure it\'s a not corrupted hdf5 file and is a real file'.format(self.fname, e)

		return messsage

	def newGroup(self, fmode='a', **kwards):
		if type(fmode) is not str or fmode not in ['a', 'w']:
			raise ValueError('unexpected value passed to fmode, expected \'a\' or \'w\', got \'{}\''.format(fmode))

		with t.open_file(self.fname, mode=fmode) as f:
			for groupName, val in kwards.items():
				f.create_earray(f.root, groupName, self.atom, (0, *val))

	def removeGroup(self, *args):
		with t.open_file(self.fname, mode='a') as f:
			for groupName in args:
				f.remove_node('/{}'.format(groupName), recursive=True)

	def renameGroup(self, **kwards):
		with t.open_file(self.fname, mode='a') as f:
			for oldName, newName in kwards.items():
				x = eval('f.root.{}'.format(oldName))
				f.rename_node(x, newName)

	def save(self, **kwards):
		with t.open_file(self.fname, mode='a') as f:
			for key, val in kwards.items():
				x = eval('f.root.{}'.format(key))
				x.append(val)

	def load(self, *args, batch_obj=':'):
		out = []
		with t.open_file(self.fname, mode='r') as f:
			for key in args:
				x = eval('f.root.{}[np.s_[{}]]'.format(key, batch_obj))
				out.append(x)

		return out

	def getShape(self, arrName):
		return self.getShapes(arrName)[0]

	def getShapes(self, *arrNames):
		out = []
		with t.open_file(self.fname, mode='r') as f:
			for i in arrNames:
				out.append(eval('f.root.{}.shape'.format(i)))

		return out

	def _iterator_init(self):
		if self.fname is None:
			raise ValueError('self.fname is empty')

		if self.iterItems is None or self.iterOutput is None:
			raise ValueError('self.iterItems and/or self.iterOutput are not defined')

		if len(self.iterItems) != 2 or len(self.iterOutput) != 2:
			raise ValueError('self.iterItems or self.iterOutput has wrong dimensions, self.iterItems has to be [[list of x array names], [list of y array names]] and self.iterOutput is the name map self.iterItems folowing the same dimensions')

		if not isinstance(self.mask_callback, type(None)) and hasattr(self.mask_callback, '__call__'):
			self._useCallback = True

		L = [i[0] for i in self.getShapes(*self.iterItems[0])]
		L.extend([i[0] for i in self.getShapes(*self.iterItems[1])])
		dataset_limit = min(L)

		index = 0

		while 1:
			self._a.append(index)
			self._b.append(index+self.batch_size)

			index += self.batch_size

			if self.limit is not None:
				if index >= self.limit or index >= dataset_limit:
					break

			elif index >= dataset_limit:
				break

	def _iterator(self):
		if self.iterItems is None or self.iterOutput is None or self.fname is None:
			raise ValueError('self.iterItems and/or self.iterOutput and/or self.fname is empty')

		if len(self.iterItems) != 2 or len(self.iterOutput) != 2:
			raise ValueError('self.iterItems or self.iterOutput has wrong dimensions, self.iterItems is [[list of x array names], [list of y array names]] and self.iterOutput is the name map for them')

		if self.shuffle:
			np.random.seed(None)
			st = np.random.get_state()
			np.random.set_state(st)
			np.random.shuffle(self._a)
			np.random.set_state(st)
			np.random.shuffle(self._b)

		index = 0

		with t.open_file(self.fname, mode='r') as f:
			while 1:
				if self.shuffle:
					np.random.seed(None)
					st = np.random.get_state()

				stepX = {}
				stepY = {}
				for name, key in zip(self.iterItems[0], self.iterOutput[0]):
					x = eval('f.root.{}[{}:{}]'.format(name, self._a[index], self._b[index]))
					if self.shuffle:
						np.random.set_state(st)
						np.random.shuffle(x)

					stepX[key] = x

				for name, key in zip(self.iterItems[1], self.iterOutput[1]):
					y = eval('f.root.{}[{}:{}]'.format(name, self._a[index], self._b[index]))
					if self.shuffle:
						np.random.set_state(st)
						np.random.shuffle(y)

					stepY[key] = y

				if self._useCallback:
					yield self.mask_callback((stepX, stepY))
				else:
					yield (stepX, stepY)

				index += 1

				if index >= len(self._a):
					index = 0
					if self.loopforever != True:
						break

					elif self.shuffle:
						np.random.seed(None)
						st = np.random.get_state()
						np.random.set_state(st)
						np.random.shuffle(self._a)
						np.random.set_state(st)
						np.random.shuffle(self._b)
		return

	@contextmanager
	def generator(self, mask_callback=None):
		try:
			self.mask_callback = mask_callback
			self._iterator_init()

			yield self._iterator

		except:
			raise

	@contextmanager
	def makeGenerator(self, layerNames, limit=None, batch_size=1, shuffle=False, mask_callback=None, **kwards):
		try:
			self.fname = './temp.h5'

			self.iterItems = layerNames
			self.iterOutput = layerNames
			self.limit = limit
			self.batch_size = batch_size
			self.shuffle = shuffle
			self.mask_callback = mask_callback

			d = {layerName: val.shape[1:] for layerName, val in kwards.items()}

			self.newGroup(fmode='w', **d)
			self.save(**kwards)

			del d
			del kwards

			self._iterator_init()

			yield self._iterator

		except:
			raise

		finally:
			if os.path.isfile('./temp.h5'):
				os.unlink('./temp.h5')


import setuptools

with open('readmePypi.md', 'r') as f:
	ld = f.read()

setuptools.setup(
	name="lose",
	version="0.6.6",
	description="A helper package for handling data in hdf5 format",
	long_description=ld,
	long_description_content_type="text/markdown",
	author="okawo",
	author_email="okawo.198@gmail.com",
	url="https://github.com/okawo80085/lose",
	packages=setuptools.find_packages(),
	install_requires=['tables', 'numpy'],
	license='MIT',
	classifiers=[
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Operating System :: OS Independent",
		"Intended Audience :: Developers",
		"Intended Audience :: Education",
		"Intended Audience :: Science/Research",
		"Topic :: Software Development :: Libraries :: Python Modules",
		"Topic :: Scientific/Engineering :: Artificial Intelligence",
	],
	python_requires='>=3.5',
)
